package com.hortonworks.spark.sql.hive.llap;

import com.google.common.base.Preconditions;
import com.hortonworks.spark.sql.hive.llap.common.HWConf;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.fs.permission.FsPermission;
import org.apache.spark.sql.Dataset;
import org.apache.spark.sql.Row;
import org.apache.spark.sql.SparkSession;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.Connection;
import java.sql.SQLException;
import java.time.Instant;
import java.time.format.DateTimeFormatter;

import static com.hortonworks.spark.sql.hive.llap.DefaultJDBCWrapper.*;
import static com.hortonworks.spark.sql.hive.llap.common.HWConf.DEFAULT_DB;
import static com.hortonworks.spark.sql.hive.llap.common.HWConf.STAGING_OUTPUT_TZ;
import static com.hortonworks.spark.sql.hive.llap.util.JobUtil.deleteDirectory;


public class SecureAccessModeExecutor {
    private static final Logger LOG = LoggerFactory.getLogger(SecureAccessModeExecutor.class);
    private static final String CONF_STRING = "set hive.exec.orc.delta.streaming.optimizations.enabled=true;" +
            "set hive.stats.autogather=false;set tez.runtime.pipelined.sorter.lazy-allocate.memory=true";

    private static final SecureAccessModeExecutor INSTANCE = new SecureAccessModeExecutor();
    private MessageDigest messageDigest = null;
    private String appLevelStagingDir = null;
    private FileSystem fs;
    private volatile boolean inited = false;
    private SparkSession sparkSession;
    private boolean unpersist = false;

    private SecureAccessModeExecutor() {
    }

    public static SecureAccessModeExecutor getInstance() {
        return INSTANCE;
    }

    public Dataset<Row> execute(HiveWarehouseSessionState sessionState, String sql) {
        Preconditions.checkNotNull(sql, "The query cannot be null");
        initIfNeeded(sessionState);
        String tableName = getTableName(sql);
        String tableLocation = getTableLocation(tableName);
        Path path = new Path(tableLocation);
        try {
            if(unpersist && fs.exists(path)){
                deleteDirectory(tableLocation,fs);
            }
            if (!fs.exists(path)){
                createDirWithPermissions(tableLocation);
                persistQueryResult(sessionState, tableName, tableLocation, sql);
            } else {
                LOG.info("Table already exists in staging space for query: {}", sql);
            }
        } catch (IOException e) {
            throw new RuntimeException(e.getMessage(), e);
        }
        return sparkSession.read().orc(tableLocation + "/*");
    }

    public void deleteStagingDir() {
        if (inited) {
            LOG.info("Deleting staging dir");
            deleteDirectory(appLevelStagingDir, fs);
        }
    }

    private synchronized void initIfNeeded(HiveWarehouseSessionState sessionState) {
        if (!inited) {
            this.sparkSession = sessionState.session;
            String sparkAppId = sparkSession.sparkContext().applicationId();
            String sparkUser = sparkSession.sparkContext().sparkUser();
            String stagingDirPrefix = HWConf.LOAD_STAGING_DIR.getString(sessionState);
            this.unpersist = sparkSession.sparkContext().conf().getBoolean("spark.hadoop.secure.access.cache.disable", false);
            LOG.debug("staging dir from conf:" + stagingDirPrefix);

            String date = Instant.ofEpochMilli(sparkSession.sparkContext().startTime())
                    .atZone(java.time.ZoneId.of(STAGING_OUTPUT_TZ.getString(sessionState)))
                    .toLocalDate().format(DateTimeFormatter.ofPattern("yyyy_MM_dd"));

            this.appLevelStagingDir = stagingDirPrefix + "/" + date + "/" + sparkUser + "/" + sparkAppId;

            try {
                Path path = new Path(appLevelStagingDir);

                Preconditions.checkNotNull(path.toUri().getScheme(),
                        "Please specify the fully qualified path of staging Dir to be used for staging output mode. " +
                        "Example: spark.datasource.hive.warehouse.load.staging.dir=s3a://<bucket-name>/staging/" +
                                " or hdfs://localhost:<port>/tmp/staging");

                this.fs = path.getFileSystem(sparkSession.sparkContext().hadoopConfiguration());
                this.messageDigest = MessageDigest.getInstance("SHA-256");
            } catch (IOException | NoSuchAlgorithmException e) {
                throw new RuntimeException(e.getMessage(), e);
            }

            this.inited = true;
        }
    }

    private String getTableName(String sql) {
        return "table_" + getQuerySha(sql);
    }

    private String getTableLocation(String tableName) {
        return appLevelStagingDir + "/" + tableName;
    }

    private String getQuerySha(String sql) {
        String querySha = null;
        try {
            querySha = String.format("%032x", new BigInteger(1,
                    messageDigest.digest(sql.getBytes("UTF-8"))));
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException(e.getMessage(), e);
        }
        Preconditions.checkNotNull(querySha, "query sha calculation failed - null generated");
        return querySha;
    }

    private void createDirWithPermissions(String location) throws IOException {
        FsPermission Permission = new FsPermission("1700");
        fs.mkdirs(new Path(location), Permission);
        LOG.debug("Creating staging table");
    }

    private void persistQueryResult(HiveWarehouseSessionState sessionState, String tableName, String tableLocation, String sql) {
        String ctas = String.format("CREATE TEMPORARY  EXTERNAL TABLE %s STORED AS ORC LOCATION '%s'  AS SELECT * FROM (%s) as q",
                tableName, tableLocation, sql);

        try (Connection conn = getConnector(sessionState)) {
            int rowsSelected = executeUpdate(conn, DEFAULT_DB.getString(sessionState), CONF_STRING, ctas);
            LOG.info("Rows Selected: " + rowsSelected);
            if (rowsSelected == 0) {
                String query = String.format("Select * from %s", tableName);
                Dataset<Row> df = executeStmt(conn, DEFAULT_DB.getString(sessionState), query, 0)
                        .asDataFrame(sessionState.session);

                df.write().orc(tableLocation + "/schema");
            }
            LOG.debug("CTAS SUCCESS!");
        } catch (SQLException e) {
            // Ctas execution failed, so we need to delete the staging dir.
            deleteDirectory(tableLocation, fs);
            throw new RuntimeException(e.getMessage(), e);
        }
    }
}

