package com.hortonworks.spark.sql.hive.llap.util;

import com.google.common.base.Preconditions;
import com.hortonworks.hwc.plan.HwcPlannerStatistics;
import com.hortonworks.spark.sql.hive.llap.DefaultJDBCWrapper;
import com.hortonworks.spark.sql.hive.llap.common.Column;
import com.hortonworks.spark.sql.hive.llap.common.DescribeTableOutput;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.mapred.JobConf;
import org.apache.spark.SparkContext;
import org.apache.spark.TaskContext;
import org.apache.spark.sql.SparkSession;
import org.apache.spark.sql.internal.SQLConf;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import scala.Option;
import scala.Tuple2;
import scala.Tuple3;
import scala.collection.Iterator;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.util.Enumeration;
import java.time.Instant;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.OptionalLong;
import java.util.TimeZone;

public class JobUtil {

  public static final String SPARK_SQL_EXECUTION_ID = "spark.sql.execution.id";
  private static Logger LOG = LoggerFactory.getLogger(JobUtil.class);
  public static final TimeZone UTC_TIME_ZONE = TimeZone.getTimeZone("UTC");

  /**
   * Gets SparkSession - getActiveSession() returns None in case of pyspark
   * falls back to getDefaultSession()
   */
  public static SparkSession getSparkSession() {
    Option<SparkSession> session = SparkSession.getActiveSession();
    if (session.isEmpty()) {
      session = SparkSession.getDefaultSession();
    }
    return session.get();
  }

  public static void populateSparkConfs(Configuration jobConf) {
    SQLConf conf = getSparkSession().sessionState().conf();
    // populate spark runtime configs

    // conf.getAllDefinedConfs() gives all the SQL confs that spark maintains the list of.
    // We need this for certain confs which are by default populated by spark e.g. 'spark.sql.session.timeZone'
    Iterator<Tuple3<String, String, String>> allConfs = conf.getAllDefinedConfs().iterator();
    while (allConfs.hasNext()) {
      Tuple3<String, String, String> keyValDesc = allConfs.next();
      jobConf.set(keyValDesc._1(), keyValDesc._2());
    }

    // conf.getAllConfs() gives the confs defined by the user
    Iterator<Tuple2<String, String>> sparkRuntimeConfIter = conf.getAllConfs().iterator();
    while (sparkRuntimeConfIter.hasNext()) {
      Tuple2<String, String> keyVal = sparkRuntimeConfIter.next();
      jobConf.set(keyVal._1(), keyVal._2());
    }
  }
  /**
   * Adjusts input millis to destination timezone.
   * <p>
   * Similar to the code in ORC project -
   * https://github.com/apache/orc/blob/rel/release-1.6.7/java/core/src/java/org/apache/orc/impl/SerializationUtils.java#L1395-L1417
   *
   * @param sourceTz      the timezone we are moving from
   * @param destinationTz the timezone we are moving to
   * @param millis        the point in time
   * @return modified milliseconds according to destination timezone
   */
  public static long convertMillis(final long millis, final TimeZone sourceTz, final TimeZone destinationTz) {
    final long sourceOffset = sourceTz.getOffset(millis);
    final long destinationOffset = destinationTz.getOffset(millis);
    final long adjustedMillis = millis + sourceOffset - destinationOffset;
    // If the timezone adjustment moves the millis across a DST boundary, we
    // need to reevaluate the offsets.
    final long dstAdjustedOffset = sourceOffset - destinationTz.getOffset(adjustedMillis);

    // final millis after adjusting the offsets
    return millis + dstAdjustedOffset;
  }

  /**
   * Adjusts input micros to destination timezone.
   * <p>
   * Similar to the code in ORC project -
   * https://github.com/apache/orc/blob/rel/release-1.6.7/java/core/src/java/org/apache/orc/impl/SerializationUtils.java#L1395-L1417
   *
   * @param sourceTz      the timezone we are moving from
   * @param destinationTz the timezone we are moving to
   * @param micros        the point in time
   * @return modified micros according to destination timezone
   */
  public static long convertMicros(final long micros, final TimeZone sourceTz, final TimeZone destinationTz) {
    final long millis = micros / 1000;
    final long remainingMicros = micros % 1000;
    return convertMillis(millis, sourceTz, destinationTz) * 1000 + remainingMicros;
  }

  public static TimeZone getSparkSessionTimezoneAtDriver() {
    return TimeZone.getTimeZone(getSparkSession().conf().get(SQLConf.SESSION_LOCAL_TIMEZONE().key()));
  }

  public static TimeZone getSparkSessionTimezoneAtExecutor(Configuration conf) {
    return TimeZone.getTimeZone(conf.get(SQLConf.SESSION_LOCAL_TIMEZONE().key()));
  }

  public static byte[] serializeJobConf(JobConf jobConf) throws IOException {
    // on local, jobConf serialized byte[] was around around 84kb
    // let's keep it 200_000 to prevent multiple array copies.
    ByteArrayOutputStream confByteArrayStream = new ByteArrayOutputStream(200_000);

    try (DataOutputStream confByteData = new DataOutputStream(confByteArrayStream)) {
      jobConf.write(confByteData);
    }
    return confByteArrayStream.toByteArray();
  }

  public static String getSqlExecutionIdAtDriver() {
    return SparkSession.getActiveSession().get().sparkContext().getLocalProperty(SPARK_SQL_EXECUTION_ID);
  }

  public static String getSqlExecutionIdAtExecutor() {
    return TaskContext.get().getLocalProperty(SPARK_SQL_EXECUTION_ID);
  }

  public static void replaceSparkHiveDriver() throws Exception {
    Enumeration<Driver> drivers = DriverManager.getDrivers();
    while (drivers.hasMoreElements()) {
      Driver driver = drivers.nextElement();
      String driverName = driver.getClass().getName();
      LOG.debug("Found a registered JDBC driver {}", driverName);
      if (driverName.endsWith("HiveDriver")) {
        LOG.debug("Deregistering {}", driverName);
        DriverManager.deregisterDriver(driver);
      } else {
        LOG.debug("Not deregistering the {}", driverName);
      }
    }
    DriverManager.registerDriver((Driver) Class.forName("org.apache.hive.jdbc.HiveDriver").newInstance());
  }

  /**
   * Makes a jdbc call and gets the output of `desc formatted {table}`. Based on that populates
   * - sizeInBytes
   * - numRows
   * in HwcPlannerStatistics.
   * Currently spark supports these two only for custom datasources.
   * <p>
   * Different datasources may use this method to get the table statistics if they are implementing {@link org.apache.spark.sql.sources.v2.reader.SupportsReportStatistics}
   *
   * @param database database name
   * @param table    table name
   * @param conn     jdbc connection instance
   * @return HwcPlannerStatistics, throws exception if numRows or totalSize is not found.
   */
  public static HwcPlannerStatistics getStatisticsForTable(String database, String table, Connection conn) {
    DescribeTableOutput describeTable = DefaultJDBCWrapper.describeTable(conn, database, table);

    String totalSize = null;
    String numRows = null;

    for (Column column : describeTable.getDetailedTableInfoColumns()) {

      if (column.getDataType() != null) {
        if (column.getDataType().trim().equals("numRows")) {
          numRows = column.getComment();
        } else if (column.getDataType().trim().equals("totalSize")) {
          totalSize = column.getComment();
        }
      }

      if (totalSize != null && numRows != null) {
        break;
      }
    }

    Preconditions.checkNotNull(numRows, "numRows found null for table - " + table);
    Preconditions.checkNotNull(totalSize, "totalSize found null for table - " + table);

    return new HwcPlannerStatistics(OptionalLong.of(Long.parseLong(totalSize)),
        OptionalLong.of(Long.parseLong(numRows)));
  }

  public static String getStagingDirPrefix(String defaultStagingDirLocation) {
    Preconditions.checkNotNull(defaultStagingDirLocation, "Staging output location is null, Enter valid value");
    SparkContext sparkContext = SparkSession.getActiveSession().get().sparkContext();
    String date = getApplicationStartDate(SparkSession.getActiveSession().get(), "UTC").format(DateTimeFormatter.ofPattern("yyyy_MM_dd"));
    return defaultStagingDirLocation + '/' + date + '/' + sparkContext.sparkUser() + '/' + sparkContext.applicationId();
  }

  public static void deleteDirectory(String deleteDir, FileSystem fs) {

    try {
      LOG.debug("deleting dir: {}", deleteDir);

      if (fs.exists(new Path(deleteDir))) {
        LOG.debug("deleting  dir: {}", deleteDir);
        fs.delete(new Path(deleteDir), true); // temp or staging  location
      }
    } catch (IOException ioException) {
      throw new RuntimeException(ioException.getMessage(), ioException);
    }

  }

  public static Long getApplicatioStartTime(SparkSession sparkSession) {
    return sparkSession.sparkContext().startTime();
  }


  public static LocalDate getApplicationStartDate(SparkSession sparkSession, String TZ) {
    Long millis = getApplicatioStartTime(sparkSession);
    return Instant.ofEpochMilli(millis).atZone(java.time.ZoneId.of(TZ)).toLocalDate();
  }
}
