package com.hortonworks.spark.sql.hive.llap.util;

import com.hortonworks.spark.sql.hive.llap.DefaultJDBCWrapper;
import com.hortonworks.spark.sql.hive.llap.common.HWConf;
import org.antlr.runtime.tree.CommonTree;
import org.apache.hadoop.hive.conf.HiveConf;
import org.apache.hadoop.hive.ql.Context;
import org.apache.hadoop.hive.ql.parse.ASTNode;
import org.apache.hadoop.hive.ql.parse.HiveParser;
import org.apache.hadoop.hive.ql.parse.ParseException;
import org.apache.hadoop.hive.ql.parse.ParseUtils;
import scala.Option;
import java.io.IOException;
import java.sql.Connection;
import java.util.Map;

import static com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil.ExecutionMethod.EXECUTE_HIVE_JDBC;
import static com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil.ExecutionMethod.EXECUTE_JDBC_CLUSTER;

public final class QueryExecutionUtil {

  public final static String HIVE_JDBC_URL_FOR_EXECUTOR = "hs2.jdbc.url.for.executor";

  public enum ExecutionMethod {
    EXECUTE_HIVE_JDBC, EXECUTE_JDBC_CLUSTER
  }

  /**
   * When smartExecutionEnabled is set, this method tries to figure out best execution method(hive jdbc or hive llap + spark data readers).
   * Currently it examines the token at root node of parse tree to separate out various patterns of select queries from others.
   * All select queries are executed via EXECUTE_QUERY_LLAP as of now.
   */
  public static ExecutionMethod resolveExecutionMethod(boolean smartExecutionEnabled,
                                                       ExecutionMethod currentExecutionMethod, String sql) {
    ExecutionMethod resolved = currentExecutionMethod;
    if (smartExecutionEnabled) {
      ASTNode tree;
      try {
        HiveConf conf = new HiveConf();
        conf.set("_hive.hdfs.session.path", "/tmp/some_dummy_path");
        conf.set("_hive.local.session.path", "/tmp/some_dummy_path");
        Context ctx = new Context(conf);
        tree = ParseUtils.parse(sql, ctx);
      } catch (ParseException | IOException e) {
        throw new RuntimeException(e.getMessage(), e);
      }
      //select queries have token of type HiveParser.TOK_QUERY at top level in their parse tree.
      if (HiveParser.TOK_QUERY == tree.getType()) {
        resolved = EXECUTE_JDBC_CLUSTER;
      } else {
        //Apart from select queries, all other queries can be executed via execute()
        resolved = EXECUTE_HIVE_JDBC;
      }
    }
    return resolved;
  }

  /*
   * Currently it examines the token at root node of parse tree to separate out various patterns of select queries from others.
   */
  public static boolean isDataFetchQuery(String sql) {
    ASTNode tree;
    try {
      HiveConf conf = new HiveConf();
      conf.set("_hive.hdfs.session.path", "/tmp/some_dummy_path");
      conf.set("_hive.local.session.path", "/tmp/some_dummy_path");
      Context ctx = new Context(conf);
      tree = ParseUtils.parse(sql, ctx);
    } catch (ParseException | IOException e) {
      throw new RuntimeException(e.getMessage(), e);
    }
    // no child
    if (tree.getChildCount() == 0)
      return false;

    //select queries have token of type HiveParser.TOK_QUERY at top level and
    // HiveParser.TOK_FROM as child in their parse tree.
    if (tree.getText().equalsIgnoreCase("TOK_QUERY")) {
      boolean containsTokFrom = tree.getChildren().stream()
              .map(n -> (ASTNode) n).map(CommonTree::getToken)
              .anyMatch(token -> token.getText().equalsIgnoreCase("TOK_FROM"));

      if (containsTokFrom)
        return true;
    }
    //Apart from select queries, all other queries can be executed via execute()
    return false;
  }

  public static Connection getConnection(Map<String, String> options) {
    String url = HWConf.RESOLVED_HS2_URL.getFromOptionsMap(options);
    String user = HWConf.USER.getFromOptionsMap(options);
    String password = HWConf.PASSWORD.getFromOptionsMap(options);
    return DefaultJDBCWrapper.getConnector(url, user, password);
  }

  public static Connection getConnection(String url, String user, String password) {
    return DefaultJDBCWrapper.getConnector(url, user, password);
  }

}
