/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.hortonworks.spark.sql.rule

import com.hortonworks.spark.sql.hive.llap.HiveWarehouseSessionImpl.{HWC_QUERY_MODE, HWC_SESSION_ID_KEY}
import com.hortonworks.spark.sql.hive.llap.common.HWConf.HWCReadMode
import com.hortonworks.spark.sql.hive.llap.common.HwcSparkListener
import com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil
import com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil.ExecutionMethod.EXECUTE_HIVE_JDBC
import com.hortonworks.spark.sql.hive.llap.{HiveWarehouseConnector, HiveWarehouseSessionImpl}
import com.qubole.spark.datasources.hiveacid.sql.execution.SparkAcidSqlParser
import com.qubole.spark.hiveacid.datasource.{HiveAcidDataSource, HiveAcidDataSourceV2}
import com.qubole.spark.hiveacid.transaction.SparkAcidQueryListener
import org.apache.log4j.{Level, LogManager, Logger}
import org.apache.spark.internal.Logging
import org.apache.spark.sql.catalyst.catalog.{CatalogTable, HiveTableRelation}
import org.apache.spark.sql.catalyst.plans.logical.{InsertIntoTable, LogicalPlan}
import org.apache.spark.sql.catalyst.rules.Rule
import org.apache.spark.sql.execution.command.DDLUtils
import org.apache.spark.sql.execution.datasources.LogicalRelation
import org.apache.spark.sql.execution.datasources.v2.DataSourceV2Relation
import org.apache.spark.sql.internal.SQLConf
import org.apache.spark.sql.{SparkSession, SparkSessionExtensions}
import java.util.regex.Pattern
import java.util.{Locale, HashMap => JHashMap, Map => JMap}


class Extensions extends Function1[SparkSessionExtensions, Unit] {

  override def apply(extensions: SparkSessionExtensions): Unit = {

    extensions.injectResolutionRule(_ => new HWCSwitchRule)
    extensions.injectParser { (session, parser) =>
      SparkAcidSqlParser(parser)
    }
  }
}

class HWCSwitchRule extends Rule[LogicalPlan] with Logging {
  val ACCESSTYPE_NONE: Byte = 1
  val ACCESSTYPE_READWRITE: Byte = 8

  private def getAccessType(tableMeta: CatalogTable): Int = {
    try {
      val method1 = tableMeta.getClass.getMethod("accessInfo")
      val accessInfo = method1.invoke(tableMeta)
      val method2 = accessInfo.getClass.getMethod("accessType")
      method2.invoke(accessInfo).asInstanceOf[Int]
    } catch {
      case n: NoSuchMethodException =>
        logInfo("Unknown access type; Assuming READWRITE")
        ACCESSTYPE_READWRITE
    }
  }

  private def getSessionOpts(keyPrefix: String, spark: SparkSession): Map[String, String] = {
    val conf = SQLConf.get

    val pattern = Pattern.compile(s"^spark\\.datasource\\.$keyPrefix\\.(.+)")
    conf.getAllConfs.flatMap { case (key, value) =>
      val m = pattern.matcher(key)
      if (m.matches() && m.groupCount() > 0) {
        Seq((m.group(1), value))
      } else {
        Seq.empty
      }
    }
  }

  val logger: Logger = LogManager.getLogger(this.getClass)
  logger.setLevel(Level.INFO)

  private def isConvertible(relation: HiveTableRelation): Boolean = {
    val serde = relation.tableMeta.storage.serde.getOrElse("").toLowerCase(Locale.ROOT)
    relation.tableMeta.properties.getOrElse("transactional", "false").toBoolean
  }

  private def isWriteEnabled(): Boolean = {
    var isWriteEnable: Boolean = false
    val spark = SparkSession.getActiveSession.get
    isWriteEnable = spark.sessionState.conf.getConfString("tests.enable.write.via.direct.reader.mode", "false") == "true"
    if (isWriteEnable == true) {
      logger.info("Write through spark-acid is enabled for testing only and is not recommended for production. [isWriteEnabled= " + isWriteEnable + "]")
    } else {
      logger.info("Write or Insert into table, through spark-acid is not supported please use HWC")
    }
    isWriteEnable
  }

  private def convert(relation: HiveTableRelation): LogicalPlan = {
    val spark = SparkSession.getActiveSession.get
    val options = relation.tableMeta.properties ++
      relation.tableMeta.storage.properties ++ Map("table" -> relation.tableMeta.qualifiedName)
    val newRelation = new HiveAcidDataSource().createRelation(spark.sqlContext, options)
    LogicalRelation(newRelation, isStreaming = false)
  }

  private def convertV2(relation: HiveTableRelation): LogicalPlan = {
    val serde = relation.tableMeta.storage.serde.getOrElse("").toLowerCase(Locale.ROOT)
    if (!serde.equals("org.apache.hadoop.hive.ql.io.orc.orcserde")) return convert(relation)
    val dbName = relation.tableMeta.identifier.database.getOrElse("default")
    val tableName = relation.tableMeta.identifier.table
    val tableOpts = Map("database" -> dbName, "table" -> tableName)
    DataSourceV2Relation.create(new HiveAcidDataSourceV2, tableOpts, None, None)
  }

  private def registerListener(): Unit = {
    if (SparkSession.active.sparkContext.getLocalProperty("hwc.listeners.registered") == null) {
      log.info("Registering Listeners")
      SparkSession.active.sparkContext.setLocalProperty("hwc.listeners.registered", "true")
      SparkSession.active.sparkContext.addSparkListener(new HwcSparkListener)
      SparkSession.active.listenerManager.register(new SparkAcidQueryListener(SparkSession.active));
    }
  }

  override def apply(plan: LogicalPlan): LogicalPlan = {
    registerListener()
    plan.resolveOperators {

    case r@HiveTableRelation(tableMeta, _, _) if (getAccessType(tableMeta) == ACCESSTYPE_NONE &&
      HWCReadMode.resolveReadMode(HWCReadMode.JDBC_CLIENT)) =>
      val dbName = tableMeta.identifier.database.getOrElse("default")
      val tableName = tableMeta.identifier.table
      log.info("using hwc extension for jdbc-client")
      // If we're using JDBC, check that the mode is 'cluster'
      val jdbcMode = HWCReadMode.JDBC_CLIENT.name()
      // Don't switch, just keep the old Hive relation
      log.warn(s"JDBC mode is $jdbcMode; Not switching to HWC connector for " +
        s"table `${dbName}`.`${tableName}`")
      return r

    case r@HiveTableRelation(tableMeta, _, _) if (getAccessType(tableMeta) == ACCESSTYPE_NONE &&
      HWCReadMode.resolveReadMode(HWCReadMode.JDBC_CLUSTER)) =>
      val dbName = tableMeta.identifier.database.getOrElse("default")
      val tableName = tableMeta.identifier.table
      log.info("using hwc extension for jdbc-cluster mode")
      val spark = SparkSession.getActiveSession.get
      val ds = new HiveWarehouseConnector
      val keyPrefix = ds.keyPrefix()
      var sessionOpts = getSessionOpts(keyPrefix, spark)
      //      // If we're using JDBC, check that the mode is 'cluster'
      //      val jdbcMode = HWConf.READ_JDBC_MODE.getFromOptionsMap(sessionOpts.asJava)

      log.info(s"Switching to Hive Warehouse Connector for table `${dbName}`.`${tableName}`")
      // get the HiveWarehouseConnector session for this spark session
      val hwcSession = HWCSwitchRule.hwcSessions.synchronized {
        val prevSession = HWCSwitchRule.hwcSessions.get(spark)
        if (prevSession != null) {
          prevSession
        } else {
          val newSession = com.hortonworks.hwc.HiveWarehouseSession.session(spark).build()
          HWCSwitchRule.hwcSessions.put(spark, newSession)
          newSession
        }
      }
      val sessionId = hwcSession.getSessionId
      log.info(s"Hive Warehouse Connector session ID is $sessionId")

      // now that we've  created the Hive Warehouse Session, get opts again (since it's filled
      // in the resolved JDBC url for us)
      sessionOpts = getSessionOpts(keyPrefix, spark)

      // The scheme for specifying the table to the HWC differs depending on whether the
      // configuration specified LLAP or JDBC. This is due to CDPD-6892
      val jdbcUrlForExecutor = hwcSession.getJdbcUrlForExecutor
      val tableOpts = Map("table" -> s"`${dbName}`.`${tableName}`",
        HWC_SESSION_ID_KEY -> sessionId,
        HWC_QUERY_MODE -> EXECUTE_HIVE_JDBC.name(),
        QueryExecutionUtil.HIVE_JDBC_URL_FOR_EXECUTOR -> jdbcUrlForExecutor)

      val options = Map(HWC_SESSION_ID_KEY -> sessionId) ++
        tableOpts ++
        sessionOpts
      DataSourceV2Relation.create(ds, options, None, None)
    // Spark-acid Write path
    case InsertIntoTable(r: HiveTableRelation, partition, query, overwrite, ifPartitionNotExists)
      if query.resolved && DDLUtils.isHiveTable(r.tableMeta) && isConvertible(r) && isWriteEnabled() &&
        HWCReadMode.resolveReadMode(HWCReadMode.DIRECT_READER_V1)=>
      log.info("using spark acid extension for writing")
      InsertIntoTable(convert(r), partition, query, overwrite, ifPartitionNotExists)

    // Spark-acid Read path - DSv1
    case relation: HiveTableRelation
      if DDLUtils.isHiveTable(relation.tableMeta) && isConvertible(relation) &&
        HWCReadMode.resolveReadMode(HWCReadMode.DIRECT_READER_V1) =>
      log.info("using DIRECT_READER_V1 extension for reading")
      convert(relation)

    // Spark-acid Read path - DSv2
    case relation: HiveTableRelation
      if DDLUtils.isHiveTable(relation.tableMeta) && isConvertible(relation) &&
        HWCReadMode.resolveReadMode(HWCReadMode.DIRECT_READER_V2) =>
      log.info("using DIRECT_READER_V2 extension for reading")
      convertV2(relation)

  }
  }
}

object HWCSwitchRule {
  private val hwcSessions: JMap[SparkSession, HiveWarehouseSessionImpl] =
    new JHashMap[SparkSession, HiveWarehouseSessionImpl]()
}
