/*
 * Copyright 2019 Qubole, Inc.  All rights reserved.
 *
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.qubole.spark.hiveacid.datasource

import com.qubole.spark.hiveacid.hive.HiveAcidMetadata
import com.qubole.spark.hiveacid.merge.{MergeWhenClause, MergeWhenNotInsert}
import com.qubole.spark.hiveacid.util.HWCDateTimeUtils.convertMillis
import com.qubole.spark.hiveacid.{HiveAcidTable, SparkAcidConf}
import org.apache.spark.internal.Logging
import org.apache.spark.rdd.RDD
import org.apache.spark.sql.catalyst.AliasIdentifier
import org.apache.spark.sql.catalyst.expressions.Expression
import org.apache.spark.sql.catalyst.util.DateTimeUtils
import org.apache.spark.sql.sources.{And, BaseRelation, EqualNullSafe, EqualTo, Filter, GreaterThan, GreaterThanOrEqual, In, InsertableRelation, LessThan, LessThanOrEqual, Not, Or, PrunedFilteredScan}
import org.apache.spark.sql.types.StructType
import org.apache.spark.sql.{Column, DataFrame, Row, SQLContext, SparkSession}
import java.sql.Timestamp
import scala.collection.JavaConversions.mapAsJavaMap

/**
  * Container for all metadata, configuration and schema to perform operations on
  * Hive ACID datasource. This provides for plumbing most of the heavy lifting is
  * performed inside HiveAcidtTable.
  *
  * @param sparkSession Spark Session object
  * @param fullyQualifiedTableName Table name for the data source.
  * @param parameters user provided parameters required for reading and writing,
  *        including configuration
  */
case class HiveAcidRelation(sparkSession: SparkSession,
                            fullyQualifiedTableName: String,
                            parameters: Map[String, String])
    extends BaseRelation
    with InsertableRelation
    with PrunedFilteredScan
    with Logging {

  private val hiveAcidMetadata: HiveAcidMetadata = HiveAcidMetadata.fromSparkSession(
    sparkSession,
    fullyQualifiedTableName
  )
  private val hiveAcidTable: HiveAcidTable = new HiveAcidTable(sparkSession,
    hiveAcidMetadata, parameters)

  def endTxn() : Unit = hiveAcidTable.endTxn()

  private val readOptions = SparkAcidConf(sparkSession, parameters)

  override def sqlContext: SQLContext = sparkSession.sqlContext

  override val schema: StructType = if (readOptions.includeRowIds) {
    hiveAcidMetadata.tableSchemaWithRowId
  } else {
    hiveAcidMetadata.tableSchema
  }

  override def insert(data: DataFrame, overwrite: Boolean): Unit = {
   // sql insert into and overwrite
    if (overwrite) {
      hiveAcidTable.insertOverwrite(data)
    } else {
      hiveAcidTable.insertInto(data)
    }
  }

  def update(condition: Option[Column], newValues: Map[String, Column]): Unit = {
    hiveAcidTable.update(condition, newValues)
  }

  def delete(condition: Column): Unit = {
    hiveAcidTable.delete(condition)
  }
  override def sizeInBytes: Long = {
    // try getting actual size of the table first
    hiveAcidMetadata.hTable.getParameters
      .getOrDefault("totalSize", sparkSession.sessionState.conf.defaultSizeInBytes.toString).toLong

  }

  def merge(sourceDf: DataFrame,
            mergeExpression: Expression,
            matchedClause: Seq[MergeWhenClause],
            notMatched: Option[MergeWhenNotInsert],
            sourceAlias: Option[AliasIdentifier],
            targetAlias: Option[AliasIdentifier]): Unit = {
    hiveAcidTable.merge(sourceDf, mergeExpression, matchedClause,
      notMatched, sourceAlias, targetAlias)
  }

  def getHiveAcidTable(): HiveAcidTable = {
    hiveAcidTable
  }

  // FIXME: should it be true / false. Recommendation seems to
  //  be to leave it as true
  override val needConversion: Boolean = false

  override def buildScan(requiredColumns: Array[String], filters: Array[Filter]): RDD[Row] = {
    val readOptions = SparkAcidConf(sparkSession, parameters)

    // since we perform a conversion of timestamps (UTC --> JVM timezone) for outputting / displaying
    // (see "case oi: TimestampObjectInspector" in HiveAcidReader.scala)
    // we have to again perform a reverse conversion(JVM --> UTC timezone) of timestamp filters to make it work correctly.
    filters.foreach(convertIfTimestampFilter)
    // sql "select *"
    hiveAcidTable.getRdd(requiredColumns, filters, readOptions)
  }

  private def convertIfTimestampFilter(filter: Filter): Unit = {
    def convertIfTimestamp(value: Any): Unit = {
      value match {
        case ts: Timestamp =>
          val nanos = ts.getNanos
          val utcMillis = convertMillis(ts.getTime, DateTimeUtils.defaultTimeZone(), DateTimeUtils.TimeZoneUTC)
          ts.setTime(utcMillis)
          ts.setNanos(nanos)
        case _ =>
      }
    }

    filter match {
      case EqualTo(_, value) => convertIfTimestamp(value)
      case EqualNullSafe(_, value) => convertIfTimestamp(value)
      case GreaterThan(_, value) => convertIfTimestamp(value)
      case GreaterThanOrEqual(_, value) => convertIfTimestamp(value)
      case LessThan(_, value) => convertIfTimestamp(value)
      case LessThanOrEqual(_, value) => convertIfTimestamp(value)
      case In(_, values) => values.foreach(convertIfTimestamp)

      case And(left, right) =>
        convertIfTimestampFilter(left)
        convertIfTimestampFilter(right)
      case Or(left, right) =>
        convertIfTimestampFilter(left)
        convertIfTimestampFilter(right)
      case Not(child) =>
        convertIfTimestampFilter(child)

      case _ =>
    }
  }

}
