package com.qubole.spark.hiveacid.reader.v2

import com.qubole.spark.hiveacid.rdd.HiveAcidPartition
import java.io.IOException
import java.util.Locale
import org.apache.hadoop.hive.ql.io.orc.OrcSplit
import org.apache.hadoop.hive.serde2.ColumnProjectionUtils
import org.apache.hadoop.mapred.{JobConf, JobID, TaskAttemptID, TaskID}
import org.apache.hadoop.mapreduce.TaskType
import org.apache.orc.{OrcConf, OrcFile, TypeDescription}
import org.apache.spark.TaskContext
import org.apache.spark.sql.catalyst.InternalRow
import org.apache.spark.sql.catalyst.analysis.caseSensitiveResolution
import org.apache.spark.sql.sources.v2.reader.InputPartitionReader
import org.apache.spark.sql.types.{StructField, StructType}
import org.apache.spark.sql.vectorized.ColumnarBatch
import scala.collection.JavaConverters.asScalaBufferConverter

private[v2] class HiveAcidInputPartitionReaderV2(split: HiveAcidPartition,
                                                 conf : JobConf,
                                                 //conf : SerializableWritable[JobConf],
                                                 partitionValues : InternalRow,
                                                 requiredFields: Array[StructField],
                                                 dataSchema: StructType,
                                                 partitionSchema : StructType,
                                                 isFullAcidTable: Boolean)
        extends InputPartitionReader[ColumnarBatch] /*with Logging */{
  //TODO : need to get a unique id and cache it.
  /*private val shouldCloneJobConf =
    sparkContext.getConf.getBoolean("spark.hadoop.cloneConf", defaultValue = false)
  private val jobConf : JobConf = HiveAcidCommon.getJobConf(broadcastedConf.value.value, null, shouldCloneJobConf)*/
  private val jobConf : JobConf = conf
  private val sparkOrcColReader : OrcColumnarBatchReader =
    new OrcColumnarBatchReader(1024)
 
  private def initReader() : Unit = {
    // Get the reader schema using the column names and types set in hive conf.
    val readerSchema: TypeDescription =
      org.apache.hadoop.hive.ql.io.orc.OrcInputFormat.getDesiredRowTypeDescr(jobConf, true, 2147483647)

    val fileSplit = split.inputSplit.value.asInstanceOf[OrcSplit]
    val readerLocal = OrcFile.createReader(fileSplit.getPath,
      OrcFile.readerOptions(jobConf).maxLength(
        OrcConf.MAX_FILE_LENGTH.getLong(jobConf)).filesystem(fileSplit.getPath.getFileSystem(jobConf)))

    // Check whether the file contains transactional information or not.
    val isAcidScan = !fileSplit.isOriginal && isFullAcidTable
    val fileSchema = if (!isAcidScan) {
      readerLocal.getSchema
    } else {
      // File will contain five transaction-related columns
      // which needs to be ignored. So get the fifth child
      // which is the reader schema.
      readerLocal.getSchema.getChildren.get(5)
    }

    // Set it as orc.mapred.input.schema so that the reader will read only the column in the file.
    jobConf.set("orc.mapred.input.schema", fileSchema.toString)

    val requestedColIds = getRefinedRequestedColIds(fileSchema)

    // Register the listener for closing the reader before init is done.
    val attemptId = new TaskAttemptID(new TaskID(new JobID(), TaskType.MAP, 0), 0)
    val taskAttemptContext = new org.apache.hadoop.mapred.TaskAttemptContextImpl(jobConf, attemptId)
    val iter = new org.apache.spark.sql.execution.datasources.RecordReaderIterator(sparkOrcColReader)
    Option(TaskContext.get()).foreach(_.addTaskCompletionListener[Unit](_ => iter.close()))

    sparkOrcColReader.initialize(fileSplit, taskAttemptContext)
    sparkOrcColReader.initBatch(readerLocal.getSchema, requestedColIds,
      requiredFields, partitionSchema, partitionValues,  isFullAcidTable, fileSplit.isOriginal)
  }
  initReader()

  @throws(classOf[IOException])
  override def next() : Boolean = {
    sparkOrcColReader.nextKeyValue()
  }

  override def get () : ColumnarBatch = {
    sparkOrcColReader.getCurrentValue
  }

  @throws(classOf[IOException])
  override def close() : Unit = {
    sparkOrcColReader.close()
  }

  def getRefinedRequestedColIds(schema: TypeDescription) : Array[Int] = {
    // Fetch Column Ids and names which were already set by Driver.
    // This information is fetched from metastore.
    val colIds = jobConf.get(ColumnProjectionUtils.READ_COLUMN_IDS_CONF_STR)
    val colNames = jobConf.get(ColumnProjectionUtils.READ_COLUMN_NAMES_CONF_STR)

    // Get field names present in the ORC file.
    val orcFieldNames = schema.getFieldNames.asScala

    val isPositionalSchema = OrcConf.FORCE_POSITIONAL_EVOLUTION.getBoolean(conf)
    val isCaseSensitive = jobConf.getBoolean("spark.sql.caseSensitive", false)

    if (!orcFieldNames.isEmpty && !colIds.isEmpty && !colNames.isEmpty) {
      // Get Column IDs array
      val requestedColIds = colIds.split(",").map(a => a.toInt)
      // Get Column Names array
      val requestedColNames = colNames.split(",")
      val mapColIdsToColNames = requestedColIds.zip(requestedColNames).toMap
      // Check whether positional schema evolution is enforced or field names begin with '_col'.
      if (isPositionalSchema || orcFieldNames.forall(_.startsWith("_col"))) {
        // Check whether the number of fields present in file
        // is more than what is present in the table.
        // If such is the case, fail read operation.
        assert(orcFieldNames.length <= dataSchema.length, "The given data schema " +
          s"${dataSchema.catalogString} has less fields than the actual ORC physical schema, " +
          "no idea which columns were dropped, fail to read.")
        // Assign index based on the position of the column (this is present in requestedColIds).
        val refinedColIds = requestedColIds.map(colId => {
          if (colId < orcFieldNames.length) {
            colId
          } else {
            -1
          }
        })

        // Set the columns which are required for reading the ORC file.
        // All indices which are -1 are filtered out.
        jobConf.set(ColumnProjectionUtils.READ_COLUMN_IDS_CONF_STR,
          refinedColIds.filter(_ != -1).sorted.mkString(","))
        jobConf.set(ColumnProjectionUtils.READ_COLUMN_NAMES_CONF_STR,
          refinedColIds.filter(_ != -1).sorted.map(colId => mapColIdsToColNames.get(colId).get).mkString(","))

        refinedColIds
      } else {
        // Name schema evolution is enforced.
        // Check whether columns are case sensitive or not.
        if (isCaseSensitive) {
          // Handling of case sensitive column names.
          val refinedColIds = requestedColIds.map(colId => {
            orcFieldNames.indexWhere(
              caseSensitiveResolution(_, mapColIdsToColNames.get(colId).get.toLowerCase(Locale.ROOT)))
          })

          // Set the columns which are required for reading the ORC file.
          // All indices which are -1 are filtered out.
          jobConf.set(ColumnProjectionUtils.READ_COLUMN_IDS_CONF_STR,
            refinedColIds.filter(_ != -1).sorted.mkString(","))
          jobConf.set(ColumnProjectionUtils.READ_COLUMN_NAMES_CONF_STR,
            refinedColIds.filter(_ != -1).sorted.map(index => orcFieldNames(index)).mkString(","))

          refinedColIds
        } else {
          // Handling of Case insensitive column names
          // Create a mapping of ORC fields names and columns
          val caseInsensitiveOrcFieldMap = orcFieldNames.zipWithIndex.groupBy(_._1.toLowerCase(Locale.ROOT))
          val refinedColIds = requestedColIds.map( index => {
            caseInsensitiveOrcFieldMap
              .get(mapColIdsToColNames.get(index).get.toLowerCase(Locale.ROOT))
              .map(matchedOrcFields => {
                if (matchedOrcFields.size > 1) {
                  // If more than 1 column has same name then throw Exception.
                  val matchedOrcFieldsString = matchedOrcFields.map(_._1).mkString("[", ", ", "]")
                  throw new RuntimeException(s"""Found duplicate field(s) "${requestedColNames(index)}": """
                    + s"$matchedOrcFieldsString in case-insensitive mode")
                } else {
                  matchedOrcFields.head._2
                }
              }).getOrElse(-1)
          })

          // Set the columns which are required for reading the ORC file.
          // All indices which are -1 are filtered out.
          jobConf.set(ColumnProjectionUtils.READ_COLUMN_IDS_CONF_STR,
            refinedColIds.filter(_ != -1).sorted.mkString(","))
          jobConf.set(ColumnProjectionUtils.READ_COLUMN_NAMES_CONF_STR,
            refinedColIds.filter(_ != -1).sorted.map(index => orcFieldNames(index)).mkString(","))

          refinedColIds
        }
      }
    } else {
      Array[Int]()
    }
  }
}
