/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.qubole.spark.hiveacid.transaction

import com.qubole.spark.hiveacid.{HiveAcidErrors, HiveAcidOperation, SparkAcidConf}
import com.qubole.spark.hiveacid.datasource.HiveAcidDataSource
import com.qubole.spark.hiveacid.hive.HiveConverter
import java.util.concurrent.atomic.AtomicBoolean
import org.apache.hadoop.hive.common.{ValidTxnList, ValidTxnWriteIdList, ValidWriteIdList}
import org.apache.hadoop.hive.metastore.{IMetaStoreClient, LockComponentBuilder, LockRequestBuilder, RetryingMetaStoreClient}
import org.apache.hadoop.hive.metastore.api.{DataOperationType, LockRequest, LockResponse, LockState, TxnInfo}
import org.apache.hadoop.hive.metastore.conf.MetastoreConf
import org.apache.hadoop.hive.metastore.txn.TxnUtils
import org.apache.spark.internal.Logging
import org.apache.spark.sql.{SparkSession, SqlUtils}
import org.apache.thrift.TException
import java.util.concurrent.{Executors, ScheduledExecutorService, ThreadFactory, TimeUnit}
import scala.collection.JavaConversions.{asScalaBuffer, asScalaSet, seqAsJavaList}
import scala.language.implicitConversions

/**
 * Txn Manager for hive acid tables.
 * This takes care of creating txns, acquiring locks and sending heartbeats
 * @param sparkSession - Spark session
 */
private[hiveacid] class HiveAcidTxnManager(sparkSession: SparkSession) extends Logging {

  private val hiveConf = HiveConverter.getHiveConf(sparkSession.sparkContext)

  private val heartbeatInterval = MetastoreConf.getTimeVar(hiveConf,
    MetastoreConf.ConfVars.TXN_TIMEOUT, TimeUnit.MILLISECONDS) / 3

  private lazy val client: IMetaStoreClient =
    RetryingMetaStoreClient.getProxy(hiveConf, false)

  private lazy val heartBeaterClient: IMetaStoreClient =
    RetryingMetaStoreClient.getProxy(hiveConf, false)

  // FIXME: Use thread pool so that we don't create multiple threads
  private val heartBeater: ScheduledExecutorService =
    Executors.newSingleThreadScheduledExecutor(new ThreadFactory() {
      def newThread(r: Runnable) = new HeartBeaterThread(r, "AcidDataSourceHeartBeater")
    })
  heartBeater.scheduleAtFixedRate(
    new HeartbeatRunnable(),
    0,
    heartbeatInterval,
    TimeUnit.MILLISECONDS)

  private val user: String = sparkSession.sparkContext.sparkUser

  private val shutdownInitiated: AtomicBoolean = new AtomicBoolean(false)
  // activeTxns is modified with synchronized in this class, so should be read same way.
  def getAllActiveTxn: List[(Long, HiveAcidTxn)] = synchronized {
    HiveAcidTxnManagerObject.activeTxns.toList
  }

  /**
    * Register transactions with Hive Metastore and tracks it under HiveAcidTxnManagerObject.activeTxns
    * @param txn transaction which needs to begin.
    * @return Update transaction object
    */
  def beginTxn(txn: HiveAcidTxn): Long = synchronized {
    // 1. Open transaction
    val txnId = client.openTxn(HiveAcidDataSource.NAME)
    if (HiveAcidTxnManagerObject.activeTxns.contains(txnId)) {
      throw HiveAcidErrors.repeatedTxnId(txnId, HiveAcidTxnManagerObject.activeTxns.keySet.toSeq)
    }
    HiveAcidTxnManagerObject.activeTxns.put(txnId, txn)
    sparkSession.conf.set("spark.acid.current.txn.id", txnId.toString)
    logDebug(s"Adding txnId: $txnId to tracker")
    txnId
  }

  /**
    * Wrapper over call to hive metastore to end transaction either with commit or abort.
    * @param txnId id of transaction to be end
    * @param abort true if transaction is to be aborted.
    */
  def endTxn(txnId: Long, abort: Boolean = false): Unit = synchronized {
    try {
      // NB: Remove it from tracking before making HMS call
      // which can potentially fail.
      HiveAcidTxnManagerObject.activeTxns.remove(txnId)
      logInfo(s"Removing txnId: $txnId from tracker")
      if (abort) {
        client.abortTxns(scala.collection.JavaConversions.seqAsJavaList(Seq(txnId)))
      } else {
        client.commitTxn(txnId)
      }
    } catch {
      case e: Exception =>
        logError(s"Failure to end txn: $txnId, presumed abort", e)
    }
  }

  /**
    * Destroy the transaction object. Closes all the pooled connection,
    * stops heartbeat and aborts all running transactions.
    */
  def close(): Unit = synchronized {

    if (!shutdownInitiated.compareAndSet(false, true)) {
      return
    }

    // Stop the heartbeat executor
    if (heartBeater != null) {
      heartBeater.shutdown()
    }

    // NB: caller of close itself is from heartbeater thread
    //    such await would be self deadlock.
    // heartBeater.awaitTermination(10, TimeUnit.SECONDS)

    // abort all active transactions
    HiveAcidTxnManagerObject.endAllTxn()

    // close all clients
    if (client != null) {
      client.close()
    }
    if (heartBeaterClient != null) {
      heartBeaterClient.close()
    }
  }

  /**
    * Returns current write id.
    * @param txnId transaction id for which current write id is requested.
    * @param dbName: Database name
    * @param tableName: Table name
    * @return
    */
  def getCurrentWriteId(txnId: Long, dbName: String, tableName: String): Long = synchronized {
    client.allocateTableWriteId(txnId, dbName, tableName)
  }

  /**
    * Return list of valid txn list.
    * @param txnIdOpt txn id, current if None is passed.
    * @return
    */
  def getValidTxns(txnIdOpt: Option[Long]): ValidTxnList = synchronized {
    txnIdOpt match {
      case Some(id) => client.getValidTxns(id)
      case None => client.getValidTxns()
    }
  }

  def getTxn(txnId: Long): HiveAcidTxn = synchronized {
    HiveAcidTxnManagerObject.activeTxns.get(txnId).orNull
  }

  /**
    * Return list of all valid write ids for the table.
    * @param fullyQualifiedTableName name of the table
    * @param validTxnList valid txn list snapshot.
    * @return List of valid write ids
    */
  def getValidWriteIds(validTxnList: ValidTxnList,
                        fullyQualifiedTableName: String): ValidWriteIdList = synchronized {
    getValidWriteIds(None, validTxnList, fullyQualifiedTableName)
  }

  /**
    * Return list of all valid write ids for the table for given transactions
    * @param txnId transaction id
    * @param validTxnList valid txn list snapshot.
    * @param fullyQualifiedTableName table name
    * @return List of valid write ids
    */
  def getValidWriteIds(txnId: Long,
                       validTxnList: ValidTxnList,
                       fullyQualifiedTableName: String): ValidWriteIdList = synchronized {
    getValidWriteIds(Option(txnId), validTxnList, fullyQualifiedTableName)
  }

  private def getValidWriteIds(txnIdOpt: Option[Long],
                               validTxnList: ValidTxnList,
                               fullyQualifiedTableName: String): ValidWriteIdList = synchronized {
    val txnId = txnIdOpt match {
      case Some(id) => id
      case None => -1L
    }
    val tableValidWriteIds = client.getValidWriteIds(Seq(fullyQualifiedTableName),
      validTxnList.writeToString())
    val txnWriteIds: ValidTxnWriteIdList = TxnUtils.createValidTxnWriteIdList(txnId,
      tableValidWriteIds)
    txnWriteIds.getTableValidWriteIdList(fullyQualifiedTableName)
  }

  def convertToDataOperationType(operationType: HiveAcidOperation.OperationType):
  DataOperationType = {
    operationType match {
      case HiveAcidOperation.INSERT_OVERWRITE => DataOperationType.UPDATE
      case HiveAcidOperation.INSERT_INTO => DataOperationType.INSERT
      case HiveAcidOperation.READ => DataOperationType.SELECT
      case HiveAcidOperation.UPDATE => DataOperationType.UPDATE
      case HiveAcidOperation.DELETE => DataOperationType.DELETE
      case _ =>
        throw HiveAcidErrors.invalidOperationType(operationType.toString)
    }
  }

  def showOpenTrans(): Seq[TxnInfo] = {
    client.showTxns().getOpen_txns.toSeq
  }

  /**
    * API to acquire locks on partitions
    * @param txnId transaction id
    * @param lockInfo: List of objects to be locked
    * @param operationType lock type
    */
  def acquireLocks(txnId: Long,
                   lockInfo: LockInfo,
                   operationType: HiveAcidOperation.OperationType,
                   conf: SparkAcidConf): Unit = synchronized {
    def createLockRequest() = {
      if (!lockInfo.getLocked) {
        val requestBuilder = new LockRequestBuilder(HiveAcidDataSource.NAME
          + " " +
          sparkSession.sparkContext.applicationId)
        requestBuilder.setUser(user)
        requestBuilder.setTransactionId(txnId)
        lockInfo.addLockComponents(requestBuilder, operationType)
        requestBuilder.build
      } else {
        null
      }
    }

    def lock(lockReq: LockRequest): Unit = {
      if (lockInfo.getLocked) {
        return
      }
      var nextSleep = 50L
      // Exponential backoff that starts with 50 milliseconds to the max of conf.maxSleepBetweenLockRetries
      // Max retries allowed are conf.lockNumRetries
      val defaultMaxSleep = conf.maxSleepBetweenLockRetries
      val MAX_SLEEP = Math.max(15000, defaultMaxSleep)
      val maxNumWaits: Int = Math.max(0, conf.lockNumRetries)
      def backoff(): Unit = {
        nextSleep *= 2
        if (nextSleep > MAX_SLEEP) nextSleep = MAX_SLEEP
        try
          Thread.sleep(nextSleep)
        catch {
          case _: InterruptedException =>

        }
      }
      try {
        var res: LockResponse = client.lock(lockReq)
        // link lockId to queryId
        var numRetries: Int = 0
        while (res.getState == LockState.WAITING && numRetries < maxNumWaits) {
          numRetries += 1
          backoff()
          res = client.checkLock(res.getLockid)
        }
        if (res.getState != LockState.ACQUIRED) {
          // Release lock in WAITING state
          if (res.getState == LockState.WAITING) {
            client.unlock(res.getLockid)
          }
          throw HiveAcidErrors.couldNotAcquireLockException(state = res.getState.name())
        }
        lockInfo.setLocked()
      } catch {
        case e: TException =>
          logWarning("Unable to acquire lock", e)
          throw HiveAcidErrors.couldNotAcquireLockException(e)
      }
    }

    lock(createLockRequest())
  }

  def addDynamicPartitions(txnId: Long, writeId: Long,
                           dbName: String,
                           tableName: String,
                           partitionNames: Set[String],
                           operationType: HiveAcidOperation.OperationType): Unit = {
    client.addDynamicPartitions(txnId, writeId, dbName,
      tableName, scala.collection.JavaConversions.seqAsJavaList(partitionNames.toSeq),
      convertToDataOperationType(operationType))
  }

  private class HeartbeatRunnable() extends Runnable {
    private def send(txn: HiveAcidTxn): Unit = {
      try {
        // Does not matter if txn is already ended
        val resp = heartBeaterClient.heartbeatTxnRange(txn.txnId, txn.txnId)
        if (resp.getAborted.nonEmpty || resp.getNosuch.nonEmpty) {
          logError(s"Heartbeat failure for transaction id: ${txn.txnId} : ${resp.toString}." +
            s"Aborting...")
        } else {
          logInfo(s"Heartbeat sent for txnId: ${txn.txnId}")
        }
      } catch {
        // No action required because if heartbeat doesn't go for some time, transaction will be
        // aborted by HMS automatically. We can abort the transaction here also if we are not
        // able to send heartbeat for some time
        case e: TException =>
          logWarning(s"Failure to heartbeat for txnId: ${txn.txnId}", e)
      }
    }

    override def run(): Unit = {
      if (shutdownInitiated.get()) {
        return
      }

      // Close the txnManager
      if (SqlUtils.hasSparkStopped(sparkSession)) {
        close()
        return
      }

      val txnList = HiveAcidTxn.txnManager.getAllActiveTxn
      txnList.foreach {
          case (_, txn) =>
            logDebug(s"Sent heartbeat for txnId: ${txn.txnId}")
            send(txn)
      }
    }
  }

  class HeartBeaterThread(val target: Runnable, val name: String) extends Thread(target, name) {
    setDaemon(true)
  }
}


private [hiveacid] class LockInfo extends Logging {
  private var isPartitionedTable: Boolean = false
  private val lockTaken: AtomicBoolean = new AtomicBoolean(false)
  private val lockInfo = new scala.collection.mutable.HashMap[String,
    scala.collection.mutable.HashMap[String, Seq[String]]]

  def setLocked() : Unit = lockTaken.set(true)

  def getLocked : Boolean = lockTaken.get()

  def addTableLock(dbName: String, tblName: String) : LockInfo = {
    if (getLocked) {
      logInfo(s"Lock acquire is already done for the txn")
      return this
    }

    var tableLock = lockInfo.get(dbName).orNull
    if (tableLock == null) {
      tableLock = new scala.collection.mutable.HashMap[String, Seq[String]]
    }
    tableLock.put(tblName, null)
    lockInfo.put(dbName, tableLock)
    this
  }

  def addPartitionLock(dbName: String, tblName: String, partitionNames: Seq[String]) : LockInfo = {
    if (getLocked) {
      logInfo(s"Lock acquire is already done for the txn")
      return this
    }
    isPartitionedTable = true
    var tableLock = lockInfo.get(dbName).orNull
    if (tableLock == null) {
      tableLock = new scala.collection.mutable.HashMap[String, Seq[String]]
      tableLock.put(tblName, partitionNames)
    } else {
      var partList = tableLock.get(tblName).orNull
      if (partList == null) {
        partList = partitionNames
      } else {
        partList ++= partitionNames
      }
      /*partitionNames.foreach(partName => {
        partList += partName
      })*/
      tableLock.put(tblName, partList)
    }
    lockInfo.put(dbName, tableLock)
    this
  }

  // Consider following sequence of event
  //  T1:   R(x)
  //  T2:   R(x)
  //  T2:   W(x)
  //  T2:   Commit
  //  T1:   W(x)
  // Because read happens with MVCC it is possible that some other transaction
  // may have come and performed write. To protect against the lost write due
  // to above sequence hive maintains write-set and abort conflict transaction
  // optimistically at the commit time.
  def addLockType(lcb: LockComponentBuilder, operationType: HiveAcidOperation.OperationType): LockComponentBuilder = {
    operationType match {
      case HiveAcidOperation.INSERT_OVERWRITE =>
        lcb.setExclusive().setOperationType(DataOperationType.UPDATE)
      case HiveAcidOperation.INSERT_INTO =>
        lcb.setSharedRead().setOperationType(DataOperationType.INSERT)
      case HiveAcidOperation.READ =>
        lcb.setSharedRead().setOperationType(DataOperationType.SELECT)
      case HiveAcidOperation.UPDATE =>
        lcb.setExclWrite().setOperationType(DataOperationType.UPDATE)
      case HiveAcidOperation.DELETE =>
        lcb.setExclWrite().setOperationType(DataOperationType.DELETE)
      case _ =>
        throw HiveAcidErrors.invalidOperationType(operationType.toString)
    }
  }

  def addLockComponents(lockRequestBuilder: LockRequestBuilder, operationType: HiveAcidOperation.OperationType) : Unit = {
    for ((dbName, tableInfo) <- lockInfo) {
      for ((tableName, partitionNames) <- tableInfo) {
        if (partitionNames == null || partitionNames.isEmpty) {
          def addDPInfoForWrites(lcb: LockComponentBuilder): LockComponentBuilder = {
            if (operationType == HiveAcidOperation.INSERT_OVERWRITE ||
              operationType == HiveAcidOperation.INSERT_INTO ||
              operationType == HiveAcidOperation.UPDATE ||
              operationType == HiveAcidOperation.DELETE) {
              lcb.setIsDynamicPartitionWrite(true)
            } else lcb
          }
          val lockCompBuilder = new LockComponentBuilder()
            .setDbName(dbName)
            .setTableName(tableName)
          // if table is partitioned but no partition names are passed
          // it means those partitions are dynamic
          if (isPartitionedTable) {
            lockRequestBuilder.
              addLockComponent(addLockType(addDPInfoForWrites(lockCompBuilder), operationType).build)
            /*val lockCompBuilder = new LockComponentBuilder()
            .setDbName(dbName)
            .setTableName(tableName)
            lockRequestBuilder.addLockComponent(addLockType(lockCompBuilder, operationType).build)*/
            logDebug(s"Added lock component for database $dbName table  $tableName")
          } else {
            lockRequestBuilder.addLockComponent(addLockType(lockCompBuilder, operationType).build)
          }
        } else {
          partitionNames.foreach(partName => {
            val lockCompBuilder = new LockComponentBuilder()
              .setPartitionName(partName)
              .setDbName(dbName)
              .setTableName(tableName)
            lockRequestBuilder.addLockComponent(addLockType(lockCompBuilder, operationType).build)
            logDebug(s"Added lock component for database $dbName table  $tableName partName $partName")
          })
        }
      }
    }
  }
}

object HiveAcidTxnManagerObject {
  // Maintain activeTxns inside txnManager instead of HiveAcidTxn
  // object for it to be accessible to back ground thread running
  // inside HiveAcidTxnManager.
  private[hiveacid] val activeTxns = new scala.collection.mutable.HashMap[Long, HiveAcidTxn]()

  def openTxn(sparkSession: SparkSession) : Unit = synchronized {
    // TODO : Need to enable it for multi table transaction support.
  }

  def commitTxn(sparkSession: SparkSession) : Unit = synchronized {
    // TODO : Need to enable it for multi table transaction support.
    endAllTxn()
  }

  def endAllTxn(txnId : Long = -1): Unit = synchronized {
    // make a copy as txn.end does a remove from activeTxns map
    val txnManager = HiveAcidTxn.txnManager
    if (txnManager != null) {
      val txnList = txnManager.getAllActiveTxn
      txnList.foreach {
        case (_, txn) =>
          if (txnId == -1 || txn.txnId == txnId) {
            txn.end()
          }
      }
    }
  }
}
